(***********************************************************************)
(*                                                                     *)
(*                              Ref-indic                              *)
(*                                                                     *)
(*                        Refinement indicators                        *)
(*                                                                     *)
(*               EPI Pomdapi, INRIA Paris - Rocquencourt               *)
(*                                                                     *)
(*  Copyright 2013-2014 INRIA.                                         *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved. This file is distributed only by permission.  *)
(*                                                                     *)
(*  Hend Ben Ameur <hbenameur@yahoo.ca>                                *)
(*  Francois Clement <Francois.Clement@inria.fr>                       *)
(*  Pierre Weis <Pierre.Weis@inria.fr>                                 *)
(*                                                                     *)
(***********************************************************************)

(* $Id: gen_identity_model.ml,v 1.8 2014-02-27 08:54:56 fclement Exp $ *)

module Nat = Unatural;;

module Cmd_opt = Ucommand_option;;

let format_sprint print ident x =
  let ppf = Format.str_formatter in
  Format.fprintf ppf "@[<hv 4>@ %s =@ %a@]" ident print x;
  Format.flush_str_formatter ()
;;

let sprint_couple print = format_sprint (Lib_print.print_Couple print print);;
let sprint_array print = format_sprint (Lib_print.print_Array print);;
let sprint_array_array print = sprint_array (Lib_print.print_Array print);;
let sprint_matrix print = format_sprint (Umatrix_print.print_t print);;

let sprint_nat_couple = sprint_couple Unatural_print.print_t;;
let sprint_nat_matrix = sprint_matrix Unatural_print.print_t;;

let sprint_float_array_array = sprint_array_array Lib_print.print_quoted_float;;
let sprint_float_matrix = sprint_matrix Lib_print.print_quoted_float;;

let sprint_color_matrix = sprint_matrix Colors_print.print_t;;

let sprint_nat = format_sprint Unatural_print.print_t;;
let sprint_float = format_sprint Lib_print.print_quoted_float;;
let sprint_string = format_sprint Lib_print.print_quoted_string;;

type image = Colors.t Umatrix.t;;

let i_qual = Upervasives.qualify "Identity";;

let i_invalid_arg = Upervasives.invalid_arg_qual i_qual;;
let i_failwith = Upervasives.failwith_qual i_qual;;

(* Abtract references. *)

let init_data_image = Umatrix.of_array_array [||];;

let { Uaref.get = get_data_image; set = set_data_image; } =
  Uaref.make_aref_unique "data image" init_data_image
;;

let default_result_image_file_name = "result_image.ppm";;

let { Uaref.get = get_result_image_file_name;
      set = set_result_image_file_name; } =
  Uaref.make_aref_single "result image file name"
    default_result_image_file_name
;;

(* Command-line options. *)

let option_result_image_file_name =
  {
   Cmd_opt.key = "-result-image-file-name";
   spec = Arg.String set_result_image_file_name;
   doc =
     Printf.sprintf
       "<fname>\n\
        \tSet result image file name to <fname>.\n\
        \tDefault is %S."
       (get_result_image_file_name ());
  }
;;

let regular_opt rco =
  Cmd_opt.with_auto_short_key
    (Cmd_opt.with_auto_gnu_key
       (Cmd_opt.of_regular rco))
;;

let user_add x = Cmd_opt.user_add (regular_opt x);;

let add_options () = user_add option_result_image_file_name;;

add_options ();;

let load_data_image fname =
  let pixmap = Ppm.load_pixmap fname in
  let idata = Umatrix.of_array_array pixmap in
  let (h, w) = Umatrix.size idata in
  if h = Nat._0_ || w = Nat._0_ then
    i_invalid_arg "load_data_image" "empty image" else
  set_data_image idata;
;;

Cmd_opt.set_usage_msg
  (Format.sprintf
     "@.%s -- Identity worker@.@.\
      Usage: %s [<option>...] <input_image_file_name>@.@.\
      Options are:@."
     Sys.argv.(0) Sys.argv.(0))
;;

Cmd_opt.set_anon_fun load_data_image;;

(* Computational server. *)

let print_image = Umatrix_print.print_t Colors_print.print_t;;
let print_data = Umatrix_print.print_t Lib_print.print_quoted_float;;

let init_data = Umatrix.make Nat._1_ Nat._1_ 0.
and init_data_width = Nat._1_
and init_data_height = Nat._1_
;;
let { Uaref.get = get_data; set = set_data; } =
  Uaref.make_aref_unique "data" init_data
and { Uaref.get = get_data_width; set = set_data_width; } =
  Uaref.make_aref_unique "data width" init_data_width
and { Uaref.get = get_data_height; set = set_data_height; } =
  Uaref.make_aref_unique "data height" init_data_height
;;

let set_data_and_size dm (ni, nj) =
  set_data_height ni;
  set_data_width nj;
  set_data dm;
;;

let string_starts_with sub s = Ustring.sub s 0 (Ustring.length sub) = sub;;

let check_initialized tname f x =
  let y =
    try f x with
    | Failure s when string_starts_with "Fatal error in Uaref" s ->
        raise (Pio.Uninitialized_worker tname) in
  y
;;

(* Cost. *)

let coarse_to_fine ppm maa =
  Umatrix.map_ij (fun i _j pp_ij -> maa.(i).(Unatural.to_int pp_ij)) ppm
;;

let add_error2 s x y = s +. (x -. y)**2.;;

let cost tname ppm maa =
  let pm = coarse_to_fine ppm maa in
  let dm = check_initialized tname get_data () in
  0.5 *. (Umatrix.fold_left2 add_error2 0. dm pm)
;;

(* Optim. *)

let check_valid_zone_numbers fun_name ppm maa =
  let valid_zone_numbers =
    Uarray_ng.array_for_all_i
      (fun i ->
        Uarray.for_all
          (fun pp_ij -> pp_ij < Uarray.length (Uarray_ng.array_get maa i)))
      (Umatrix.to_array_array ppm) in
  if not valid_zone_numbers then
    i_invalid_arg fun_name "zone numbers in partition are not all valid"
;;

let _sprint_dm dm i j =
  Printf.sprintf "i = %d, j = %d, dm = %F" i j (Umatrix.get dm i j)
;;
let _sprint_maa_opt_count maa_opt count i zn =
  Printf.sprintf "i = %d, zn = %d; maa_opt = %F, count = %d"
    i zn maa_opt.(i).(zn) !(count.(i).(zn))
;;

let optim { Udebug.spf_debug = debug; _; } (ppm, maa_init) =
  debug (sprint_nat_matrix "ppm" ppm);
  debug (sprint_float_array_array "maa_init" maa_init);
  let dim = Uarray.length maa_init in
  let dim' = Umatrix.dim_i ppm in
  if dim' <> dim then assert false else begin
  check_valid_zone_numbers "optim" ppm maa_init;
  let maa_opt = Uarray_ng.array_array_map (fun _ -> 0.) maa_init
  and count = Uarray_ng.array_array_map (fun _ -> ref 0) maa_init in
  let dm = get_data () in
  if Umatrix.size dm <> Umatrix.size ppm then assert false else
  let add i j zn =
    let zn = Unatural.to_int zn in
    maa_opt.(i).(zn) <- maa_opt.(i).(zn) +. Umatrix.get dm i j;
    incr count.(i).(zn); in
  Umatrix.iter_ij add ppm;
  let maa_opt =
    Uarray_ng.array_array_map2
      (fun mi ci -> mi /. float_of_int !ci)
      maa_opt count in
  debug (sprint_float_array_array "maa_opt" maa_opt);
  let j_opt = cost "optim" ppm maa_opt in
  debug (sprint_float "j_opt" j_opt);
  (maa_opt, j_opt) end
;;

(* Init. *)

let check_dimension get_dim dim' =
  let dim = get_dim () in
  if dim' <> dim then
    invalid_arg (Printf.sprintf "dimension must be %d" (Unatural.to_int dim))
;;

let make_init get_dim dm_of_img ({ Udebug.spf_debug = debug; _; } as spf) dim =
  debug (sprint_nat "dim" dim);
  check_dimension get_dim dim;
  let img = get_data_image () in
  debug (sprint_color_matrix "img" img);
  let sz = Umatrix.size img in
  debug (sprint_nat_couple "sz" sz);
  let dm = dm_of_img img in
  let (dim', nbpt) = Umatrix.size dm in
  if dim' <> dim then
    failwith "argument image is not of the declared dimension" else
  debug (sprint_float_matrix "dm" dm);
  set_data_and_size dm sz;
  let ppm1 = Umatrix.make dim nbpt Unatural._0_ in
  debug (sprint_nat_matrix "ppm1" ppm1);
  let maa1_init = Uarray.init dim (fun _i -> Uarray.make Unatural._1_ 0.) in
  let (maa1, j1) = optim spf (ppm1, maa1_init) in
  debug (sprint_float_array_array "maa1" maa1);
  debug (sprint_float "j1" j1);
  (sz, ppm1, maa1, j1)
;;

(* Grad. *)

let grad { Udebug.spf_debug = debug; _; } pm =
  debug (sprint_float_matrix "pm" pm);
  let dm = check_initialized "grad" get_data () in
  debug (sprint_float_matrix "dm" dm);
  let gm = Umatrix.map2 ( -. ) pm dm in
  debug (sprint_float_matrix "gm" gm);
  gm
;;

(* Final. *)

let save_result_parameter
    img_of_pm { Udebug.spf_debug = debug; _; } (dname, pm) =
  debug (sprint_string "dname" dname);
  debug (sprint_float_matrix "pm" pm);
  let ni = check_initialized "final" get_data_height ()
  and nj = check_initialized "final" get_data_width () in
  let img = img_of_pm (ni, nj) pm in
  let fname = get_result_image_file_name () in
  Ppm.save_pixmap (Filename.concat dname fname) (Umatrix.to_array_array img);
;;

let make_final = save_result_parameter;;
